import asyncio
import functools
import threading

from . import aio


class BaseProxy:

    def __init__(self, ref):
        self._ref = ref


class EventLoop(threading.Thread):

    def __init__(self, name='AIOTH', loop=None):
        self.loop = loop or asyncio.new_event_loop()
        self.proxies = {}
        super().__init__(name=name)
        self.daemon = True

    def run(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def stop(self):
        self.loop.call_soon_threadsafe(self.loop.stop)

    def call_soon(self, func, *args, **kwargs):
        f = functools.partial(func, *args, **kwargs)
        return self.loop.call_soon_threadsafe(f)

    def run_coroutine(self, coro):
         return asyncio.run_coroutine_threadsafe(coro, self.loop)

    def _create_coroutine_threadsafe(self, corof, resolve_future):
        @functools.wraps(corof)
        def wrapper(obj, *args, **kwargs):
            coro = corof(obj._ref, *args, **kwargs)
            future = self.run_coroutine(coro)
            return future.result() if resolve_future else future
        return wrapper

    def _create_proxy_for(self, klass, resolve_futures=True):
        class Proxy(BaseProxy):
            pass
        for name in dir(klass):
            if name.startswith('_'):
                continue
            member = getattr(klass, name)
            if asyncio.iscoroutinefunction(member):
                member = self._create_coroutine_threadsafe(
                    member, resolve_futures)
            setattr(Proxy, name, member)
        return Proxy

    def proxy(self, obj, resolve_futures=True):
        if not self.is_alive():
            self.start()
        klass = type(obj)
        key = klass, resolve_futures
        Proxy = self.proxies.get(key)
        if not Proxy:
            Proxy = self._create_proxy_for(klass, resolve_futures)
            self.proxies[key] = Proxy
        return Proxy(obj)

    def socket(self, host, port, eol=b'\n', auto_reconnect=True,
               resolve_futures=True):
        sock = aio.Socket(host, port, eol=eol, auto_reconnect=auto_reconnect)
        return self.proxy(sock, resolve_futures)


DefaultEventLoop = EventLoop()
Socket = DefaultEventLoop.socket





import asyncio
import functools


def ensure_connection(f):
    @functools.wraps(f)
    async def wrapper(self, *args, **kwargs):
        if self.auto_reconnect and not self.connected:
            await self.open()
        return await f(self, *args, **kwargs)
    return wrapper


class Socket:

    def __init__(self, host, port, auto_reconnect=True):
        self.host = host
        self.port = port
        self._reader = None
        self._writer = None
        self.auto_reconnect = auto_reconnect

    async def open(self):
        if self.connected:
            raise ConnectionError('socket already open. must close it first')
        self._reader, self._writer = await asyncio.open_connection(
            self.host, self.port)

    async def close(self):
        if self._writer is not None:
            self._writer.close()
            await self._writer.wait_closed()
        self._reader = None
        self._writer = None

    @property
    def connected(self):
        if self._reader is None:
            return False
        eof = self._reader.at_eof()
        return not eof

    @ensure_connection
    async def read(self, n=-1):
        return await self._reader.read(n)

    @ensure_connection
    async def readline(self):
        return await self._reader.readline()

    @ensure_connection
    async def readlines(self, n):
        for i in range(n):
            yield await self._reader.readline()

    @ensure_connection
    async def readexactly(self, n):
        return await self._reader.readexactly(n)

    @ensure_connection
    async def readuntil(self, separator=b'\n'):
        await self._ensure_connected()
        return await self._reader.readuntil(separator)

    @ensure_connection
    async def write(self, data):
        self._writer.write(data)
        await self._writer.drain()

    @ensure_connection
    async def writelines(self, lines):
        self._writer.writelines(lines)
        await self._writer.drain()

    @ensure_connection
    async def write_readline(self, data):
        self._writer.write(data)
        await self._writer.drain()
        return await self._reader.readline()

    @ensure_connection
    async def write_readlines(self, data, n):
        self._writer.write(data)
        await self._writer.drain()
        for i in range(n):
            yield await self._reader.readline()

    @ensure_connection
    async def writelines_readlines(self, lines, n=None):
        if n is None:
            n = len(lines)
        self._writer.writelines(lines)
        await self._writer.drain()
        for i in range(n):
            yield await self._reader.readline()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from setuptools import setup, find_packages


if sys.version_info < (3, 5):
    print('sockio needs python >= 3.5')
    exit(1)

TESTING = any(x in sys.argv for x in ["test", "pytest"])

setup_requirements = []
if TESTING:
    setup_requirements += ['pytest-runner']
test_requirements = ['pytest', 'pytest-cov', 'pytest-asyncio']


setup(
    name='sockio',
    author="Jose Tiago Macara Coutinho",
    author_email='coutinhotiago@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    description="Concurrency agnostic socket API",
    license="MIT license",
    long_description="Concurrency agnostic socket API",
    keywords='socket, asyncio',
    packages=find_packages(include=['sockio']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/tiagocoutinho/sockio',
    version='0.1.0',
    zip_safe=True
)
